#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase
from collections import namedtuple

from schemaconvertor.convertor import (
    Schema, SchemaConst, ObjAsDictAdapter, SchemaConvertor)

Pair = namedtuple("Pair", ["key", "value"])


class TestObjAsDictAdapter(TestCase):
    class Parent(object):
        pclsval = 1

        def __init__(self):
            self.pinsval = 2

    class Child(Parent):
        cclsval = 3

        def __init__(self):
            super(TestObjAsDictAdapter.Child, self).__init__()
            self.cinsval = 4

        def __getattr__(self, name):
            if name == "dynamicval":
                return 5

    def test_base_method(self):
        parent = self.Parent()
        parent_dict = ObjAsDictAdapter(parent)

        self.assertEqual(parent_dict["pclsval"], parent.pclsval)
        self.assertEqual(parent_dict["pinsval"], parent.pinsval)
        self.assertIn("pclsval", iter(parent_dict))
        self.assertIn("pinsval", iter(parent_dict))

    def test_inherit(self):
        child = self.Child()
        child_dict = ObjAsDictAdapter(child)

        self.assertEqual(child_dict["pclsval"], child.pclsval)
        self.assertEqual(child_dict["pinsval"], child.pinsval)
        self.assertEqual(child_dict["cclsval"], child.cclsval)
        self.assertEqual(child_dict["cinsval"], child.cinsval)
        self.assertIn("pclsval", iter(child_dict))
        self.assertIn("pinsval", iter(child_dict))
        self.assertIn("cclsval", iter(child_dict))
        self.assertIn("cinsval", iter(child_dict))

    def test_dynamic_attr(self):
        child = self.Child()
        child_dict = ObjAsDictAdapter(child)

        self.assertEqual(child_dict["dynamicval"], child.dynamicval)
        self.assertNotIn("dynamicval", iter(child_dict))

        child.dynamicval = 6
        self.assertEqual(child_dict["dynamicval"], child.dynamicval)
        self.assertIn("dynamicval", iter(child_dict))

    def test_dict_interface(self):
        child = self.Child()
        parent = self.Parent()
        child_dict = ObjAsDictAdapter(child)
        parent_dict = ObjAsDictAdapter(parent)

        self.assertEqual(parent_dict.get("cinsval"), None)
        self.assertEqual(
            parent_dict.get("cinsval", child_dict["cinsval"]),
            child_dict.get("cinsval"))

        self.assertGreater(len(parent_dict), 0)
        self.assertEqual(parent_dict.keys(), list(parent_dict.iterkeys()))
        self.assertEqual(parent_dict.values(), list(parent_dict.itervalues()))
        self.assertEqual(parent_dict.items(), list(parent_dict.iteritems()))
        self.assertEqual(len(parent_dict), len(list(parent_dict)))


class TestSchema(TestCase):
    def test_field_default_value(self):
        schema = Schema({})
        self.assertEqual(schema.version, schema.VERSION)
        self.assertEqual(schema.origin_schema, {})
        self.assertEqual(schema.parent, None)
        self.assertEqual(schema.compiled, False)  # lazy
        self.assertEqual(schema.type, SchemaConst.S_UNDEFINED)
        self.assertEqual(schema.description, str(schema.__class__))
        self.assertEqual(schema.items, SchemaConst.S_DISABLED)
        self.assertEqual(schema.properties_schemas, SchemaConst.S_DISABLED)
        self.assertEqual(schema.pattern_properties_schemas, SchemaConst.S_DISABLED)
        self.assertEqual(schema.typeof_schemas, SchemaConst.S_DISABLED)
        self.assertEqual(schema.typeof_default_schema, SchemaConst.S_DISABLED)
        self.assertEqual(schema.compiled, True)
        self.assertEqual(schema.encoding, SchemaConst.V_ENCODING)
        self.assertEqual(schema.decoderrors, SchemaConst.V_DECODERR)

    def test_version(self):
        schema = Schema({
            "type": "string"
        })
        self.assertEqual(schema.version, Schema.VERSION)
        self.assertTrue(schema.check_version())

        schema = Schema({
            "version": Schema.VERSION,
            "type": "string"
        })
        self.assertEqual(schema.version, Schema.VERSION)
        self.assertTrue(schema.check_version())

        schema = Schema({
            "version": "0.2",
            "type": "string"
        })
        self.assertNotEqual(schema.version, Schema.VERSION)
        self.assertTrue(schema.check_version())

        schema = Schema({
            "version": "0.0.0.0",
            "type": "string"
        })
        self.assertEqual(schema.version, "0.0.0.0")
        self.assertFalse(schema.check_version())

        schema = Schema({
            "version": "0.0.0.0",
            "type": "array",
            "items": {
                "type": "string"
            }
        })
        self.assertEqual(schema.version, "0.0.0.0")
        self.assertEqual(schema.items.version, "0.0.0.0")

    def test_parent(self):
        schema = Schema({
            "version": "0.2",
            "type": "array",
            "items": {
                "type": "string"
            }
        })
        self.assertIs(schema.items.parent, schema)
        self.assertIs(schema.parent, None)

    def test_short(self):
        schema1 = Schema({
            "version": "0.2",
            "type": "array",
            "items": "string"
        })
        schema2 = Schema({
            "version": "0.2",
            "type": "array",
            "items": {
                "type": "string"
            }
        })
        self.assertEqual(schema1.items.type, schema2.items.type)

        schema = Schema("string")
        self.assertEqual(schema.type, "string")

    def test_encoding(self):
        schema = Schema({
            "version": "0.2",
            "type": "array",
            "items": {
                "type": "string"
            }
        })
        self.assertEqual(schema.encoding, "utf-8")
        self.assertEqual(schema.items.encoding, "utf-8")
        self.assertEqual(schema.decoderrors, "strict")
        self.assertEqual(schema.items.decoderrors, "strict")

        schema = Schema({
            "version": "0.2",
            "encoding": "gbk",
            "decoderrors": "ignore",
            "type": "array",
            "items": {
                "type": "string"
            }
        })
        self.assertEqual(schema.encoding, "gbk")
        self.assertEqual(schema.items.encoding, "gbk")
        self.assertEqual(schema.decoderrors, "ignore")
        self.assertEqual(schema.items.decoderrors, "ignore")

        schema = Schema({
            "version": "0.2",
            "encoding": "gbk",
            "decoderrors": "ignore",
            "type": "array",
            "items": {
                "type": "string",
                "encoding": "utf-8",
                "decoderrors": "replace",
            }
        })
        self.assertEqual(schema.encoding, "gbk")
        self.assertEqual(schema.items.encoding, "utf-8")
        self.assertEqual(schema.decoderrors, "ignore")
        self.assertEqual(schema.items.decoderrors, "replace")


class TestSchemaConvertor(TestCase):
    def test_string_encoding(self):
        convertor = SchemaConvertor({
            "type": "string"
        })
        self.assertEqual(convertor(u"刘奕聪"), u"刘奕聪")
        self.assertEqual(convertor(u"刘奕聪".encode("utf-8")), u"刘奕聪")

        convertor = SchemaConvertor({
            "encoding": "utf-8",
            "type": "string"
        })
        self.assertEqual(convertor(u"刘奕聪"), u"刘奕聪")
        self.assertEqual(convertor(u"刘奕聪".encode("utf-8")), u"刘奕聪")

        convertor = SchemaConvertor({
            "encoding": None,
            "type": "string"
        })
        self.assertEqual(convertor(u"刘奕聪"), u"刘奕聪")
        self.assertEqual(convertor(
            u"刘奕聪".encode("utf-8")), u"刘奕聪".encode("utf-8"))
        self.assertEqual(convertor(None), None)

        convertor = SchemaConvertor({
            "encoding": "gbk",
            "type": "string"
        })
        self.assertEqual(convertor(u"刘奕聪"), u"刘奕聪")
        self.assertEqual(convertor(u"刘奕聪".encode("gbk")), u"刘奕聪")

    def test_string_decoderr(self):
        convertor = SchemaConvertor({
            "encoding": "utf-8",
            "type": "string"
        })

        with self.assertRaises(UnicodeDecodeError):
            self.assertEqual(convertor(u"刘奕聪".encode("gbk")), u"刘奕聪")

        convertor = SchemaConvertor({
            "encoding": "utf-8",
            "decoderrors": "ignore",
            "type": "string"
        })
        self.assertIsInstance(convertor(u"刘奕聪".encode("gbk")), unicode)
        self.assertNotEqual(convertor(u"刘奕聪".encode("gbk")), u"刘奕聪")

        convertor = SchemaConvertor({
            "encoding": "utf-8",
            "decoderrors": "replace",
            "type": "string"
        })
        self.assertIsInstance(convertor(u"刘奕聪".encode("gbk")), unicode)
        self.assertNotEqual(convertor(u"刘奕聪".encode("gbk")), u"刘奕聪")

    def test_integer(self):
        convertor = SchemaConvertor({
            "type": "integer"
        })

        self.assertIs(convertor(1), 1)
        self.assertIs(convertor("2"), 2)
        self.assertIs(convertor(3.4), 3)

    def test_float(self):
        convertor = SchemaConvertor({
            "type": "float"
        })

        self.assertEqual(convertor("1.2"), 1.2)
        self.assertEqual(convertor(3.4), 3.4)

    def test_bool(self):
        convertor = SchemaConvertor({
            "type": "boolean"
        })
        self.assertTrue(convertor(not 0))
        self.assertTrue(convertor("1"))
        self.assertTrue(convertor(2.3))
        self.assertTrue(convertor([4]))
        self.assertTrue(convertor({5: 6}))
        self.assertTrue(convertor(object()))
        self.assertTrue(convertor(True))
        self.assertFalse(convertor(0))
        self.assertFalse(convertor(""))
        self.assertFalse(convertor(0.0))
        self.assertFalse(convertor([]))
        self.assertFalse(convertor({}))
        self.assertFalse(convertor(None))
        self.assertFalse(convertor(False))

    def test_number(self):
        convertor = SchemaConvertor({
            "type": "number"
        })
        self.assertEqual(convertor(1), 1)
        self.assertEqual(convertor("2.3"), 2.3)
        self.assertEqual(convertor(3.4), 3.4)
        self.assertEqual(convertor("5"), 5)

    def test_dict(self):
        convertor = SchemaConvertor({
            "type": "dict",
            "properties": {
                "key": "string",
                "value": "string"
            }
        })

        data = {
            "key": "test",
            "value": 1
        }
        self.assertDictEqual(convertor(data), {
            "key": "test",
            "value": "1"
        })

        convertor = SchemaConvertor({
            "type": "dict",
            "patternProperties": {
                "[a-z]": "string",
                "[0-9]": "number"
            }
        })

        data = {
            "a": 0,
            "1": "2",
            "b": "3.4",
            "5": 5.6,
        }
        self.assertDictEqual(convertor(data), {
            "a": "0",
            "1": 2,
            "b": "3.4",
            "5": 5.6,
        })

    def test_object(self):
        convertor = SchemaConvertor({
            "type": "object",
            "properties": {
                "key": "string",
                "value": "string"
            }
        })

        data = Pair("test", 1)
        self.assertDictEqual(convertor(data), {
            "key": "test",
            "value": "1"
        })

        convertor = SchemaConvertor({
            "type": "object",
            "patternProperties": {
                r"^k\w+$": "string",
                r"^v\w+$": "number"
            }
        })
        data = Pair("test", 1)
        self.assertDictEqual(convertor(data), {
            "key": "test",
            "value": 1
        })

    def test_array(self):
        convertor = SchemaConvertor({
            "type": "array",
            "items": "string"
        })
        data = range(3)
        self.assertListEqual(convertor(data), ["0", "1", "2"])

    def test_null(self):
        convertor = SchemaConvertor({
            "type": "null",
        })
        self.assertIsNone(convertor(None))
        self.assertIsNone(convertor(0))
        self.assertIsNone(convertor(1.2))
        self.assertIsNone(convertor("3.4"))

    def test_raw(self):
        convertor = SchemaConvertor({
            "type": "raw",
        })
        data = [None, 0, 1.2, "3"]
        self.assertListEqual(convertor(data), data)

    def test_typeof(self):
        convertor = SchemaConvertor({
            "typeOf": {
                int: "boolean",
                None: "string",
                (float, str): "integer",
                Pair: {
                    "type": "object",
                    "properties": {
                        "key": "string",
                        "value": "string"
                    }
                },
                "default": "string"
            }
        })
        self.assertEqual(convertor(1), True)
        self.assertEqual(convertor(2.3), 2)
        self.assertEqual(convertor("4"), 4)
        self.assertEqual(convertor(None), "None")
        self.assertEqual(convertor(Pair("test", 1)), {
            "key": "test", "value": "1"
        })
        self.assertEqual(convertor(True), True)  # bool is subclass of int
        self.assertEqual(convertor([]), "[]")



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase
from collections import namedtuple

from schemaconvertor.convertor import convert_by_schema


Tag = namedtuple("Tag", ["name", "value"])


class User(object):
    Role = "Normal"

    def __init__(self, name, email):
        self.name = name
        self.email = email


class Admin(User):
    Role = "Admin"

    def __init__(self, name, email, wid):
        self.wid = wid
        super(Admin, self).__init__(name, email)


class Book(object):
    def __init__(self, name, owners=()):
        self.owners = list(owners)
        self.name = name
        self.tags = {}

SIMPLEUSERSCHEMA = {
    "version": "0.3",
    "encoding": "utf-8",
    "type": "object",
    "properties": {
        "name": "string",
    }
}

FULLUSERSCHEMA = {
    "version": "0.3",
    "encoding": "utf-8",
    "type": "object",
    "properties": {
        "name": "string",
        "email": "string",
        "Role": "string",
    }
}

SIMPLEADMINSCHEMA = {
    "version": "0.3",
    "encoding": "utf-8",
    "type": "object",
    "properties": {
        "name": "string",
        "wid": "integer"
    }
}

FULLADMINSCHEMA = {
    "version": "0.3",
    "encoding": "utf-8",
    "type": "object",
    "properties": {
        "name": "string",
        "email": "string",
        "wid": "integer",
        "Role": "string",
    }
}

USERMIXTYPELISTSCHEMA = {
    "version": "0.3",
    "encoding": "utf-8",
    "type": "array",
    "items": {
        "typeOf": {
            User: SIMPLEUSERSCHEMA,
            Admin: SIMPLEADMINSCHEMA
        }
    }
}

USERLISTSCHEMA = {
    "version": "0.3",
    "encoding": "utf-8",
    "type": "array",
    "items": SIMPLEUSERSCHEMA
}

SIMPLEBOOKSCHEMA = {
    "version": "0.3",
    "encoding": "utf-8",
    "type": "object",
    "properties": {
        "name": "string",
        "owners": {
            "type": "array",
            "items": SIMPLEUSERSCHEMA
        }
    }
}

FULLBOOKSCHEMA = {
    "version": "0.3",
    "encoding": "utf-8",
    "type": "object",
    "properties": {
        "name": "string",
        "owners": {
            "type": "array",
            "items": FULLUSERSCHEMA
        },
        "tags": {
            "type": "dict",
            "patternProperties": {
                "^[a-z]+$": {
                    "type": "object",
                    "properties": {
                        "name": "string",
                        "value": "string"
                    }
                },
                r"^\w+_cnt$": "integer",
                r"^\w+_val$": "number",
            }
        }
    }
}


class TestUser(TestCase):
    def test_convert_User(self):
        user = User("lyc", "imyikong@gmail.com")
        self.assertDictEqual(convert_by_schema(user, FULLUSERSCHEMA), {
            "name": "lyc",
            "email": "imyikong@gmail.com",
            "Role": "Normal"
        })

    def test_convert_Admin(self):
        admin = Admin("刘奕聪", "imyikong@gmail.com", 0)
        self.assertDictEqual(convert_by_schema(admin, FULLADMINSCHEMA), {
            "name": u"刘奕聪",
            "email": "imyikong@gmail.com",
            "wid": 0,
            "Role": "Admin"
        })

    def test_convert_User_list(self):
        user_lst = [
            User("u1", "xxx"),
            Admin("a1", "xxx", 1),
            User("u2", "xxx"),
        ]

        self.assertLessEqual(convert_by_schema(user_lst, USERMIXTYPELISTSCHEMA), [
            {"name": "u1"},
            {"name": "a1", "wid": 1},
            {"name": "u2"},
        ])


class TestBook(TestCase):
    def test_convert_Book(self):
        user1 = User("u1", "xxx")
        admin1 = Admin("a1", "xxx", 2)
        book = Book("b1", [user1, admin1])

        self.assertDictEqual(convert_by_schema(book, SIMPLEBOOKSCHEMA), {
            "name": "b1",
            "owners": [
                {"name": "u1"},
                {"name": "a1"}
            ]
        })

    def test_convert_Book_full(self):
        user1 = User("u1", "xxx")
        admin1 = Admin("a1", "xxx", 2)
        book = Book("b1", [user1, admin1])
        book.tags["visited_cnt"] = 56
        book.tags["owners_val"] = 2
        book.tags["attr"] = Tag("public", "yes")

        self.assertDictEqual(convert_by_schema(book, FULLBOOKSCHEMA), {
            "name": "b1",
            "owners": [
                {
                    "name": "u1",
                    "email": "xxx",
                    "Role": "Normal"
                },
                {
                    "name": "a1",
                    "email": "xxx",
                    "Role": "Admin"
                }
            ],
            "tags": {
                "visited_cnt": 56,
                "owners_val": 2,
                "attr": {
                    "name": "public",
                    "value": "yes"
                }
            }
        })



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase
from collections import namedtuple

from schemaconvertor import convertor

Pair = namedtuple("Pair", ["key", "value"])


class TestSimple(TestCase):
    def test_base_type(self):
        data = {
            "string": "1",
            "integer": 2,
            "float": 3.4,
            "number_integer": 5,
            "number_float": 6.7,
            "number_str": "8.9",
            "number_asinteger": 10.0,
            "boolean": True,
            "null": None,
        }

        schema = {
            "type": "dict",
            "properties": {
                "string": {
                    "type": "string"
                },
                "integer": {
                    "type": "integer"
                },
                "float": {
                    "type": "float"
                },
                "number_integer": {
                    "type": "number"
                },
                "number_float": {
                    "type": "number"
                },
                "number_str": {
                    "type": "number"
                },
                "number_asinteger": {
                    "type": "number"
                },
                "boolean": {
                    "type": "boolean"
                },
                "null": {
                    "type": "null"
                },
            }
        }
        cvtr = convertor.SchemaConvertor(schema)
        result = cvtr(data)
        self.assertDictEqual(result, {
            "string": "1",
            "integer": 2,
            "float": 3.4,
            "number_integer": 5,
            "number_float": 6.7,
            "number_str": 8.9,
            "number_asinteger": 10.0,
            "boolean": True,
            "null": None,
        })

        schema = {
            "type": "dict",
            "properties": {
                "string": "string",
                "integer": "integer",
                "float": "float",
                "number_integer": "number",
                "number_float": "number",
                "number_str": "number",
                "number_asinteger": "number",
                "boolean": "boolean",
                "null": "null",
            }
        }
        cvtr = convertor.SchemaConvertor(schema)
        result = cvtr(data)
        self.assertDictEqual(result, {
            "string": "1",
            "integer": 2,
            "float": 3.4,
            "number_integer": 5,
            "number_float": 6.7,
            "number_str": 8.9,
            "number_asinteger": 10.0,
            "boolean": True,
            "null": None,
        })

        schema = {
            "type": "dict",
            "properties": {
                "string": "integer",
                "float": "integer",
            }
        }
        cvtr = convertor.SchemaConvertor(schema)
        result = cvtr(data)
        self.assertDictEqual(result, {
            "string": 1,
            "float": 3,
        })

    def test_raw_type(self):
        schema = {"type": "raw"}
        cvtr = convertor.SchemaConvertor(schema)

        self.assertEqual(cvtr(1), 1)
        self.assertEqual(cvtr(2.3), 2.3)
        self.assertEqual(cvtr("345"), "345")
        self.assertEqual(cvtr([6, 7.8]), [6, 7.8])
        self.assertEqual(cvtr({9: 0}), {9: 0})

    def test_complex_type(self):
        data = Pair(key="test", value="test value")
        schema = {
            "type": "object",
            "properties": {
                "key": "string",
                "value": "string"
            }
        }
        cvtr = convertor.SchemaConvertor(schema)
        self.assertEqual(cvtr(data), {
            "key": "test",
            "value": "test value"
        })

        schema = {
            "type": "array",
            "items": "string"
        }
        cvtr = convertor.SchemaConvertor(schema)
        self.assertEqual(cvtr(data), ["test", "test value"])

        data = Pair(key=1, value={
            "array": [Pair(key=2, value=3), Pair(key=4, value=5)],
            "dict": {"key": 6, "value": 7},
            "object": Pair(key=8, value=9),
            "int": 10,
        })
        schema = {
            "type": "object",
            "properties": {
                "key": "string",
                "value": {
                    "type": "dict",
                    "properties": {
                        "array": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "properties": {
                                    "key": "string",
                                    "value": "integer"
                                }
                            }
                        },
                        "dict": {
                            "type": "dict",
                            "properties": {
                                "key": "string",
                                "value": "integer"
                            }
                        },
                        "object": {
                            "type": "object",
                            "properties": {
                                "key": "string",
                                "value": "integer"
                            }
                        },
                        "int": "integer"
                    }
                }
            }
        }
        cvtr = convertor.SchemaConvertor(schema)
        self.assertEqual(cvtr(data), {
            "key": "1",
            "value": {
                "array": [
                    {"key": "2", "value": 3},
                    {"key": "4", "value": 5},
                ],
                "dict": {"key": "6", "value": 7},
                "object": {"key": "8", "value": 9},
                "int": 10
            }
        })

    def test_mixing_type(self):
        data = [
            Pair(key=1, value=2),
            3, 4.5, "6"
        ]
        schema = {
            "type": "array",
            "items": {
                "typeOf": {
                    (int, str): "string",
                    float: "integer",
                    Pair: {
                        "type": "object",
                        "properties": {
                            "key": "string",
                            "value": "integer"
                        }
                    }
                }
            }
        }
        cvtr = convertor.SchemaConvertor(schema)
        self.assertEqual(cvtr(data), [
            {"key": "1", "value": 2},
            "3", 4, "6"
        ])

        schema = {
            "type": "array",
            "items": {
                "typeOf": {
                    "default": "string",
                    Pair: {
                        "type": "object",
                        "properties": {
                            "key": "string",
                            "value": "integer"
                        }
                    }
                }
            }
        }
        cvtr = convertor.SchemaConvertor(schema)
        self.assertEqual(cvtr(data), [
            {"key": "1", "value": 2},
            "3", "4.5", "6"
        ])

        schema = {
            "type": "object",
            "properties": {
                "key": "string",
                "value": {
                    "typeOf": {
                        int: "float",
                        str: "integer",
                        float: {
                            "type": "string"
                        }
                    }
                }
            }
        }
        cvtr = convertor.SchemaConvertor(schema)
        self.assertEqual(cvtr(Pair("test", 1)), {'key': 'test', 'value': 1.0})
        self.assertEqual(cvtr(Pair("test", "2")), {'key': 'test', 'value': 2})
        self.assertEqual(cvtr(Pair("test", 3.4)), {'key': 'test', 'value': "3.4"})



#!/usr/bin/env python
# encoding: utf-8

from . import convertor

__author__ = 'Liu Yicong'
__email__ = 'imyikong@gmail.com'
__version__ = convertor.__version__
__all__ = ["__author__", "__email__", "__version__", "convertor"]



#!/usr/bin/env python
# encoding: utf-8

import types
import re
import collections

__version__ = '0.3'


def _type_convertor(type_):
    """Type convertor builder
    """
    def _base_convertor(self, data, schema):
        """Convert data to given type
        """
        return type_(data)
    return _base_convertor


SchemaVersionError = type("SchemaVersionError", (ValueError,), {})
FieldTypeError = type("FieldTypeError", (TypeError,), {})
FieldMissError = type("FieldMissError", (KeyError,), {})


class ObjAsDictAdapter(collections.Mapping):
    def __init__(self, obj):
        self.__object = obj

    def __getitem__(self, name):
        try:
            return getattr(self.__object, name)
        except AttributeError:
            raise KeyError(name)

    def __setitem__(self, name, value):
        setattr(self.__object, name, value)

    def __iter__(self):
        return iter(dir(self.__object))

    def __len__(self):
        return len(dir(self.__object))


class SchemaConst(object):
    # schema types
    T_STR = "string"
    T_OBJ = "object"
    T_INT = "integer"
    T_FLOAT = "float"
    T_NUM = "number"
    T_BOOL = "boolean"
    T_DICT = "dict"
    T_LIST = "array"
    T_NULL = "null"
    T_RAW = "raw"
    T_DEFAULT = T_STR

    # schema fields
    F_TYPE = "type"
    F_VERSION = "version"
    F_DEFAULT = "default"
    F_PROPERTIES = "properties"
    F_ITEMS = "items"
    F_TYPEOF = "typeOf"
    F_PATTERNPROPERTIES = "patternProperties"
    F_ENCODING = "encoding"
    F_DECODERR = "decoderrors"
    F_DESCRIPTION = "description"

    # field states
    S_UNDEFINED = None
    S_DISABLED = frozenset()

    # const values
    V_ENCODING = "utf-8"
    V_DECODERR = "strict"


class Schema(object):
    VERSION = __version__
    VERVERIFYREX = re.compile(r"0\.[1-3].*")

    def __init__(self, schema, parent=None):
        if isinstance(schema, basestring):
            schema = {"type": schema}

        self.origin_schema = schema
        self.parent = parent
        self.version = schema.get(
            SchemaConst.F_VERSION,
            parent.version if parent else self.VERSION)
        self.description = str(schema.get(
            SchemaConst.F_DESCRIPTION,
            parent.description if parent else self.__class__))
        self.compiled = False

    def __getattr__(self, name):
        self.compile()
        return getattr(self, name)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.origin_schema)

    def __str__(self):
        return "<Schema: %s>" % self.description

    def compile(self):
        """compile schema
        """
        if self.compiled:
            return

        schema = self.origin_schema
        self.type = schema.get(SchemaConst.F_TYPE, SchemaConst.S_UNDEFINED)

        items = schema.get(SchemaConst.F_ITEMS)
        self.items = self.subschema(items) if items else SchemaConst.S_DISABLED

        properties = schema.get(SchemaConst.F_PROPERTIES)
        self.properties_schemas = SchemaConst.S_DISABLED \
            if properties is None else {
                k: self.subschema(s) for k, s in properties.iteritems()
            }

        typeof_schemas = schema.get(SchemaConst.F_TYPEOF)
        self.typeof_schemas = SchemaConst.S_DISABLED \
            if typeof_schemas is None else {
                types.NoneType if t is None else t: self.subschema(s)
                for t, s in typeof_schemas.iteritems()
                if isinstance(t, (type, tuple, types.NoneType))
            }
        self.typeof_default_schema = SchemaConst.S_DISABLED \
            if typeof_schemas is None else self.subschema(typeof_schemas.get(
                SchemaConst.F_DEFAULT, SchemaConst.T_DEFAULT))

        p_schemas = schema.get(SchemaConst.F_PATTERNPROPERTIES)
        self.pattern_properties_schemas = SchemaConst.S_DISABLED \
            if p_schemas is None else {
                re.compile(p): self.subschema(s)
                for p, s in p_schemas.iteritems()
            }

        self.encoding = schema.get(
            SchemaConst.F_ENCODING,
            self.parent.encoding if self.parent else SchemaConst.V_ENCODING)
        self.decoderrors = schema.get(
            SchemaConst.F_DECODERR,
            self.parent.decoderrors if self.parent else SchemaConst.V_DECODERR)

        self.compiled = True

    def check_version(self):
        """Check version if is available
        """
        return self.VERVERIFYREX.match(self.version) is not None

    def properties(self, name, istry=False):
        """Get sub shcema by name
        """
        if self.properties_schemas is SchemaConst.S_DISABLED:
            return SchemaConst.S_UNDEFINED

        if name in self.properties_schemas:
            return self.properties_schemas.get(name, SchemaConst.S_UNDEFINED)
        if not istry:
            raise FieldMissError(
                "field %s is miss in %s" % (name, SchemaConst.F_PROPERTIES))

    def typeof(self, data, istry=False):
        """Get sub schema by data type
        """
        if self.typeof_schemas is SchemaConst.S_DISABLED:
            if istry:
                return SchemaConst.S_UNDEFINED
            raise FieldMissError("field %s is miss" % SchemaConst.F_TYPEOF)

        sch = self.typeof_schemas.get(type(data), SchemaConst.S_UNDEFINED)
        if sch is not SchemaConst.S_UNDEFINED:
            return sch

        for typ, sch in self.typeof_schemas.iteritems():
            if isinstance(data, typ):
                return sch
        return self.typeof_default_schema

    def pattern_properties(self, name, istry=False):
        """Get sub schema by name pattern
        """
        if self.pattern_properties_schemas is SchemaConst.S_DISABLED:
            if istry:
                return SchemaConst.S_UNDEFINED
            raise FieldMissError(
                "field %s is miss" % SchemaConst.F_PATTERNPROPERTIES)

        for rex, sch in self.pattern_properties_schemas.iteritems():
            if rex.search(name):
                return sch
        return SchemaConst.S_UNDEFINED

    def subschema(self, sch):
        """create a subschema
        """
        return Schema(sch, self)


class SchemaConvertor(object):
    def __init__(self, schema):
        if not isinstance(schema, Schema):
            schema = Schema(schema)

        if schema.check_version() is False:
            raise SchemaVersionError()

        self.schema = schema

    def __call__(self, data):
        return self._convertor(data, self.schema)

    def _convertor(self, data, schema):
        """Main convertor
        """
        convertor = self.CONVERTORS.get(schema.type)
        if convertor is None:
            raise TypeError("Unknown type: %s" % schema.type)

        return convertor(self, data, schema)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self.schema))

    def __str__(self):
        return "<SchemaConvertor: %s>" % self.schema

    def _dict_convertor(self, data, schema):
        """Dict convertor
        """
        result = {}
        if schema.pattern_properties_schemas is not SchemaConst.S_DISABLED:
            for key in data:
                real_data = data[key]
                real_schema = schema.pattern_properties(key, istry=True)
                if real_schema:
                    result[key] = self._convertor(real_data, real_schema)

        if schema.properties_schemas is not SchemaConst.S_DISABLED:
            for key in schema.properties_schemas:
                real_data = data[key]
                real_schema = schema.properties(key)
                if real_schema:
                    result[key] = self._convertor(real_data, real_schema)

        return result

    def _object_convertor(self, data, schema):
        """Object convertor
        """
        return self._dict_convertor(ObjAsDictAdapter(data), schema)

    def _array_convertor(self, data, schema):
        """iterable object convertor
        """
        result = []
        real_schema = schema.items
        if real_schema is not SchemaConst.S_DISABLED:
            for item in data:
                result.append(self._convertor(item, real_schema))
        return result

    def _number_convertor(self, data, schema):
        """Auto number convertor
        """
        num = float(data)
        if num.is_integer():
            num = int(num)
        return num

    def _null_convertor(self, data, schema):
        """Return None forever
        """
        return None

    def _raw_convertor(self, data, schema):
        """Return the raw object forever
        """
        return data

    def _auto_type_convertor(self, data, schema):
        """when schema.type is None
        """
        real_schema = schema.typeof(data, istry=True)
        if real_schema is not SchemaConst.S_UNDEFINED:
            return self._convertor(data, real_schema)
        return self._null_convertor(data, schema)

    def _str_convertor(self, data, schema):
        """auto unicode string convertor
        """
        if schema.encoding is None:
            return data

        if isinstance(data, unicode):
            return data

        data = str(data)
        return data.decode(schema.encoding, schema.decoderrors)

    CONVERTORS = {
        SchemaConst.T_STR: _str_convertor,
        SchemaConst.T_INT: _type_convertor(types.IntType),
        SchemaConst.T_FLOAT: _type_convertor(types.FloatType),
        SchemaConst.T_BOOL: _type_convertor(types.BooleanType),
        SchemaConst.T_NUM: _number_convertor,
        SchemaConst.T_DICT: _dict_convertor,
        SchemaConst.T_OBJ: _object_convertor,
        SchemaConst.T_LIST: _array_convertor,
        SchemaConst.T_NULL: _null_convertor,
        SchemaConst.T_RAW: _raw_convertor,
        None: _auto_type_convertor,
    }


def convert_by_schema(data, schema):
    """a quick tool to convert data by schema
    """
    cvtr = SchemaConvertor(schema)
    return cvtr(data)



#!/usr/bin/env python
# coding=utf-8

import os
import sys

from schemaconvertor import __version__

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


from setuptools import find_packages


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

setup(
    name='schemaconvertor',
    version=__version__,
    description='a tool to serialize objects by schema',
    long_description=open('README.rst', 'a+').read(),
    author='Liu Yicong',
    author_email='imyikong@gmail.com',
    packages=find_packages(),
    install_requires=(),
    license='BSD',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
)
